package ca.recrown.islandsurvivalcraft.interaction.items;

import org.bukkit.Material;
import org.bukkit.NamespacedKey;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.HumanEntity;
import org.bukkit.inventory.EquipmentSlot;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.Recipe;

import ca.recrown.islandsurvivalcraft.IslandSurvivalCraft;

public interface VariedItem {
    /**
     * Should be used to perform anything that need only be called once.
     * Is called only when the needInitialization method returns true.
     * Is called after assigning identity.
     * @param islandSurvivalCraft The IslandSurvivalCraft plugin object that contains references to various managers that may be used.
     */
    public void initialize(IslandSurvivalCraft islandSurvivalCraft);

    /**
     * @return the recipe to be registered with bukkit.
     */
    public Recipe getRecipe();

    /**
     * @return The official, displayed name of the item.
     */
    public String getName();

    /**
     * 
     * @return The material item this item is a variation of.
     */
    public Material[] getBaseItems();

    /**
     * @return The namespaced key Bukkit will use to identify this variation of the item.
     */
    public NamespacedKey getNamespacedKey();

    /**
     * Assigns a persistent metadata identifier.
     * This identifier should be saved to any itemstack that is considered this item.
     * @param identifier The identifier.
     */
    public void assignIdentifier(VariationItemIdentifier identifier);


    /**
     * Should give the identifier used identify this variation of the base item from the original.
     * This usually should just return the assigned identifier.
     * @return The identifier metadata.
     */
    public VariationItemIdentifier getIdentifier();

    /**
     * If this item needs initializing.
     * If returning true, the initialize method will be called before being registered as a recipe.
     * @return True if the initialize method should be called.
     */
    public boolean needInitialization();

    /**
     * Called when this variety of the item is crafted.
     * @param item The resulting itemstack.
     * @param crafter The entity performing the crafting.
     * @param recipe The recipe used to craft this.
     * @return true if the remaining chain of events for this crafting should not occur. Otherwise, false.
     */
    public boolean onItemCrafted(ItemStack item, HumanEntity crafter);
    
    /**
     * Called when this item is right clicked with.
     * @param item The itemstack that was right clicked with.
     * @param clicker The player that did the clicking.
     * @param hand The hand that it was in.
     * @param block The block the item was right clicked on.
     * @param blockFace The face of the block that was clicked.
     * @return True if the vanilla chain of events should no longer run. False otherwise.
     */
    public boolean onItemRightClicked(ItemStack item, HumanEntity clicker, EquipmentSlot hand, Block block, BlockFace blockFace);

    /**
     * Called once when the player changes to holding the item.
     * @param heldItem The itemstack representing the held item.
     * @param holder The player holding the item.
     */
    public void onItemHeld(ItemStack heldItem, HumanEntity holder);

    /**
     * Called when this item is put away.
     * @param item The itemstack that was put away.
     * @param holder The player that put away the itemstack.
     */
    public void onItemPutAway(ItemStack item, HumanEntity holder);
    
    /**
     * Called when a player joins and called when player picks up this variation of the item.
     * Can be used to initialize any data the item doesn't save on server shutdown.
     * You may need to check if the item already has been instantiated.
     * @param item The itemstack that represents the item in question.
     */
    public void initializeInstanceOfItem(ItemStack item);

    /**
     * 
     * @return A short description in regards to this variation of the item.
     */
    public String getDescription();
}