package ca.recrown.islandsurvivalcraft.world.Information;

import java.util.HashSet;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;

import ca.recrown.islandsurvivalcraft.utilities.GeneralUtilities;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMap;

public class IslandInformationManager {
    private final IslandWorldMap islandMap;
    private final ExecutorService executors = Executors.newFixedThreadPool(2, new ThreadFactory() {
        @Override
        public Thread newThread(Runnable r) {
            Thread thread = new Thread(r);
            thread.setName("ISC-Chunk-Info-Loader");
            return thread;
        }
    });
    private final ConcurrentHashMap<Point2, Future<HashSet<IslandInformation>>> building = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Point2, HashSet<IslandInformation>> islandSets = new ConcurrentHashMap<>();

    public IslandInformationManager(IslandWorldMap islandWorldMap) {
        this.islandMap = islandWorldMap;
    }

    public HashSet<IslandInformation> loadChunkIslandInformation(Point2 chunkCoords) {
        HashSet<IslandInformation> result = new HashSet<>();
        Point2[] surroundings = new Point2[4];
        surroundings[0] = chunkCoords.shift(1, 0);
        surroundings[1] = chunkCoords.shift(-1, 0);
        surroundings[2] = chunkCoords.shift(0, -1);
        surroundings[3] = chunkCoords.shift(0, 1);
        HashSet<Point2> checkedCoordinates = new HashSet<>();
        for (Point2 surrounding : surroundings) {
            if (Thread.currentThread().isInterrupted()) return null;
            HashSet<IslandInformation> chunkIslandSet = islandSets.get(surrounding);
            if (chunkIslandSet != null) {
                for (IslandInformation islandInformation : chunkIslandSet) {
                    if (islandInformation.isIslandInChunk(chunkCoords)) {
                        checkedCoordinates.addAll(islandInformation.getIslandCoordinates());
                        result.add(islandInformation);
                    }
                }
            }
        }

        for (int x = 0; x < GeneralUtilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < GeneralUtilities.CHUNK_SIZE; z++) {
                if (Thread.currentThread().isInterrupted()) return null;
                int worldX = chunkCoords.x * GeneralUtilities.CHUNK_SIZE + x;
                int worldZ = chunkCoords.y * GeneralUtilities.CHUNK_SIZE + z;
                if (islandMap.isIsland(worldX, worldZ)) {
                    IslandInformationBuilder infoBuilder = new IslandInformationBuilder();
                    Point2 origin = islandMap.findIslandOrigin(worldX, worldZ, (p) -> {
                        if (!checkedCoordinates.add(p))
                            return true;
                        infoBuilder.addCoordinate(p, islandMap.isEdgeOfIsland(p.x, p.y));
                        infoBuilder.addChunk(GeneralUtilities.worldToChunkCoordinates(p));
                        return false;
                    });
                    if (origin != null) {
                        IslandInformation islandInfo = infoBuilder.build(origin);
                        result.add(islandInfo);
                    }
                }
            }
        }
        return result;
    }

    public void loadChunkIslandInformationAsync(Point2 chunkCoords) {
        CompletableFuture<HashSet<IslandInformation>> completableFuture = new CompletableFuture<>();
        executors.execute(() -> {
            completableFuture.complete(loadChunkIslandInformation(chunkCoords));
        });
        building.put(chunkCoords, completableFuture);
        completableFuture.thenAccept((p) -> {
            if (Thread.currentThread().isInterrupted()) return;
            islandSets.put(chunkCoords, p);
            building.remove(chunkCoords);
        });
    }

    public void unloadChunkIslandInformation(Point2 chunkCoords) {
        if (building.containsKey(chunkCoords)) building.remove(chunkCoords).cancel(true);
        islandSets.remove(chunkCoords);
    }

    private HashSet<IslandInformation> getChunkIslandInformationSet(Point2 chunkCoords, boolean checkLoading) {
        if (!islandSets.containsKey(chunkCoords)) {
            if (checkLoading && building.containsKey(chunkCoords)) {
                try {
                    building.get(chunkCoords).get();
                } catch (InterruptedException | ExecutionException e) {
                }
            }
        }
        return islandSets.get(chunkCoords);
    }

    /**
     * Gets the island information for the given coordinates.
     * @param coords The coordinates to check for island information.
     * @param checkLoading Whether or not to check for the currently loading chunks.
     * @return The island information or null if there is no island at those coordinates.
     */
    public IslandInformation getIslandInformation(Point2 coords, boolean checkLoading) {
        HashSet<IslandInformation> chunkIslandSet = getChunkIslandInformationSet(GeneralUtilities.worldToChunkCoordinates(coords), checkLoading);
        if (chunkIslandSet != null) {
            for (IslandInformation islandInformation : chunkIslandSet) {
                if (islandInformation.isWithinIsland(coords)) return islandInformation;
            }
        }
        return null;
    }
}